/*  kglcylinder.cpp - 
    This file is part of KGL
    Copyright (C) 2000  Antonio Larrosa Jimenez
    KGL' homepage : http://www.arrakis.es/~rlarrosa/kgl.html

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 

    Send comments and bug fixes to Antonio Larrosa <larrosa@kde.org>

 ***************************************************************************/
#include "kglcylinder.h"
#include <math.h>
#include <stdio.h>

KGLCylinder::KGLCylinder(GLfloat x, GLfloat y, GLfloat z, GLfloat r1, GLfloat r2, GLfloat h, int slices, QImage texture1, QImage texture2) : KGLObject(x,y,z)
{
  m_list = glGenLists(1);
  glNewList(m_list, GL_COMPILE);
  GLUquadricObj* q = gluNewQuadric();
  gluQuadricDrawStyle( q, (GLenum) GLU_FILL );
  gluQuadricNormals( q, (GLenum) GLU_SMOOTH );

  if (!texture2.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    texture2 = texture2.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture2.width(), texture2.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture2.bits() );
  } else glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color() );
  gluQuadricNormals(q, GL_SMOOTH);
  gluCylinder(q, r1, r2, h, slices, 1);

  // end cap

  if (!texture1.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, glcolor(Qt::white) );
    texture1 = texture1.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture1.width(), texture1.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture1.bits() );
  } else {
    gluQuadricTexture(q, GL_FALSE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color() );
  }
  glTranslatef(0.0, 0.0, 2.0);
  gluDisk(q, 0.0, r1, slices, 1);

  // other end cap

  glTranslatef(0.0, 0.0, -2.0);

  gluQuadricOrientation(q, (GLenum)GLU_INSIDE);
  gluDisk(q, 0.0, r2, slices, 1);
  gluQuadricOrientation(q, (GLenum)GLU_OUTSIDE);
  glEndList();

  gluDeleteQuadric(q);

}

KGLCylinder::KGLCylinder(GLfloat x1, GLfloat y1, GLfloat z1, GLfloat x2, 
    GLfloat y2, GLfloat z2, GLfloat r, GLfloat r2, int slices, QImage texture1, QImage texture2) 
: KGLObject(x1,y1,z1)
{
  GLfloat h=sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)+(z2-z1)*(z2-z1));
  printf("h %g\n",h);

  
/* Let's calculate the angle between (0,0,0)(0,0,h) and (0,0,0)(x2-x1,y2-y1,z2-z1)
 * We need first to get the orthogonal projection of (0,0,h) onto the line 
 * generated by the vector (x2-x1,y2-y1,z2-z1) 
 *
 * We have to get a point (x,y,z) which verifies:
 *  x-x1   y-y2    z-z2
 *  ---- = ----  = ----
 *  x2-x1  y2-y1   z2-z1
 *
 * and
 *
 * (a-x)*(x1-x)+(b-y)*(y1-y)+(c-z)*(z1-z)=0
 *
 * That is, (x,y,z) which verifies :
 *
 *    x      y
 *  ---- = ----
 *  x2-x1  y2-y1
 *
 *    y       z
 *  ----  = ----
 *  y2-y1   z2-z1
 *
 * and
 *
 *  2   2
 * x  +y  +z*(z-h)=0
 *
 * Therefore, we must have:
 *
 *             (z2-z1)*h*(y2-y1)
 * y =  -------------------------------------
 *                2           2           2
 *         (x2-x1)   + (y2-y1)   + (z2-z1)
 *
 * and:
 *      y*(x2-x1)       y*(z2-z1)
 *  x = ---------   z = ---------
 *        y2-y1           y2-y1
 */

 GLfloat dx=x2-x1;
 GLfloat dy=y2-y1;
 GLfloat dz=z2-z1;

 printf("d %g\t%g\t%g\n",dx,dy,dz);

 GLfloat py= (dz*h*dy)/(dx*dx+dy*dy+dz*dz);
 GLfloat px= (py==0) ? 0 : py*dx/dy;
 GLfloat pz= (py==0) ? 0 : py*dz/dy;

 printf("p %g\t%g\t%g\n",px,py,pz);

 GLfloat dpx=sqrt(px*px+py*py+pz*pz);
 GLfloat dpy=sqrt(px*px+py*py+(h-pz)*(h-pz));

 
// rho is the angle between (0,0,h) and (dx,dy,dz)
 GLfloat rho=(dpy==0) ? 0 : atan2(dpy,dpx);
 
 printf("rho %g (%g %g)\n",rho,dpx,dpy);
// v is the vectorial product of (0,0,h) and d
 GLfloat vx=-h*dy;
 GLfloat vy=h*dx;
 GLfloat vz=0;
 
 printf("v %g\t%g\t%g\n",vx,vy,vz);

  m_list = glGenLists(1);
  glNewList(m_list, GL_COMPILE);
  GLUquadricObj* q = gluNewQuadric();
  gluQuadricDrawStyle( q, (GLenum) GLU_FILL );
  gluQuadricNormals( q, (GLenum) GLU_SMOOTH );

  glPushMatrix();
  glMatrixMode(GL_MODELVIEW);

  if (!texture2.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    texture2 = texture2.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture2.width(), texture2.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture2.bits() );
  } else glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color() );
  gluQuadricNormals(q, GL_SMOOTH);

  glRotatef(rho*360.0/(2*M_PI),vx,vy,vz);
  
  gluCylinder(q, r, r2, h, slices, 1);

  // end cap

  if (!texture1.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, glcolor(Qt::white) );
    texture1 = texture1.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture1.width(), texture1.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture1.bits() );
  } else {
    gluQuadricTexture(q, GL_FALSE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color() );
  }
  glTranslatef(0.0, 0.0, h);
  gluDisk(q, 0.0, r2, slices, 1);

  // other end cap

  glTranslatef(0.0, 0.0, -h);

  gluQuadricOrientation(q, (GLenum)GLU_INSIDE);
  gluDisk(q, 0.0, r, slices, 1);
  gluQuadricOrientation(q, (GLenum)GLU_OUTSIDE);
//  glRotatef(-rho,vx,vy,vz);
  glPopMatrix();
  glEndList();

  gluDeleteQuadric(q);


}

KGLCylinder::KGLCylinder(int, GLfloat x1, GLfloat y1, GLfloat z1, GLfloat phi, 
    GLfloat psi, GLfloat r, GLfloat r2, GLfloat h, int slices, QImage texture1, QImage texture2) 
: KGLObject(x1,y1,z1)
{
//  GLfloat h=sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)+(z2-z1)*(z2-z1));
  printf("h %g\n",h);

  
/* Let's calculate the angle between (0,0,0)(0,0,h) and (0,0,0)(x2-x1,y2-y1,z2-z1)
 * We need first to get the orthogonal projection of (0,0,h) onto the line 
 * generated by the vector (x2-x1,y2-y1,z2-z1) 
 *
 * We have to get a point (x,y,z) which verifies:
 *  x-x1   y-y2    z-z2
 *  ---- = ----  = ----
 *  x2-x1  y2-y1   z2-z1
 *
 * and
 *
 * (a-x)*(x1-x)+(b-y)*(y1-y)+(c-z)*(z1-z)=0
 *
 * That is, (x,y,z) which verifies :
 *
 *    x      y
 *  ---- = ----
 *  x2-x1  y2-y1
 *
 *    y       z
 *  ----  = ----
 *  y2-y1   z2-z1
 *
 * and
 *
 *  2   2
 * x  +y  +z*(z-h)=0
 *
 * Therefore, we must have:
 *
 *             (z2-z1)*h*(y2-y1)
 * y =  -------------------------------------
 *                2           2           2
 *         (x2-x1)   + (y2-y1)   + (z2-z1)
 *
 * and:
 *      y*(x2-x1)       y*(z2-z1)
 *  x = ---------   z = ---------
 *        y2-y1           y2-y1
 */
/*
 GLfloat dx=x2-x1;
 GLfloat dy=y2-y1;
 GLfloat dz=z2-z1;

 printf("d %g\t%g\t%g\n",dx,dy,dz);

 GLfloat py= (dz*h*dy)/(dx*dx+dy*dy+dz*dz);
 GLfloat px= (py==0) ? 0 : py*dx/dy;
 GLfloat pz= (py==0) ? 0 : py*dz/dy;

 printf("p %g\t%g\t%g\n",px,py,pz);

 GLfloat dpx=sqrt(px*px+py*py+pz*pz);
 GLfloat dpy=sqrt(px*px+py*py+(h-pz)*(h-pz));

 
// rho is the angle between (0,0,h) and (dx,dy,dz)
 GLfloat rho=(dpy==0) ? 0 : atan2(dpy,dpx);
 
 printf("rho %g\n",rho);
// v is the vectorial product of (0,0,h) and d
 GLfloat vx=h*dy;
 GLfloat vy=h*dx;
 GLfloat vz=0;
 
 printf("v %g\t%g\t%g\n",vx,vy,vz);
*/
  m_list = glGenLists(1);
  glNewList(m_list, GL_COMPILE);
  GLUquadricObj* q = gluNewQuadric();
  gluQuadricDrawStyle( q, (GLenum) GLU_FILL );
  gluQuadricNormals( q, (GLenum) GLU_SMOOTH );

//  glPushMatrix();
//  glMatrixMode(GL_MODELVIEW);

  if (!texture2.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    texture2 = texture2.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture2.width(), texture2.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture2.bits() );
  } else glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color() );
  gluQuadricNormals(q, GL_SMOOTH);

//  glRotatef(rho,vx,vy,vz);
  glRotatef(phi,1,0,0);
  glRotatef(psi,0,1,0);
  
  gluCylinder(q, r, r2, h, slices, 1);

  // end cap

  if (!texture1.isNull()) 
  {
    gluQuadricTexture(q, GL_TRUE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, glcolor(Qt::white) );
    texture1 = texture1.convertDepth( 32 ).mirror().swapRGB(); 
    glTexImage2D( GL_TEXTURE_2D, 0, 3, texture1.width(), texture1.height(), 0,
	GL_RGBA, GL_UNSIGNED_BYTE, texture1.bits() );
  } else {
    gluQuadricTexture(q, GL_FALSE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color() );
  }
  glTranslatef(0.0, 0.0, h);
  gluDisk(q, 0.0, r2, slices, 1);

  // other end cap

  glTranslatef(0.0, 0.0, -h);

  gluQuadricOrientation(q, (GLenum)GLU_INSIDE);
  gluDisk(q, 0.0, r, slices, 1);
  gluQuadricOrientation(q, (GLenum)GLU_OUTSIDE);
//  glRotatef(-rho,vx,vy,vz);
//  glPopMatrix();
  glEndList();

  gluDeleteQuadric(q);


}
